﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.ServiceModel;
using System.Linq;
using System.Reflection;
using System.ComponentModel;
using System.Collections.Generic;

namespace AsyncAndSync
{
    public class ServiceProxy<T> where T : ICommunicationObject
    {
        private static string assemblyName;

        private static string AssemblyName
        {
            get
            {
                if (assemblyName == null)
                    assemblyName = typeof(T).Assembly.FullName;
                return assemblyName;
            }
        }

        private static MethodInfo completeEventHandlerMethodInfo;

        private static MethodInfo CompleteEventHandlerMethodInfo
        {
            get
            {
                if (completeEventHandlerMethodInfo == null)
                {
                    var currentType = typeof(ServiceProxy<T>);
                    completeEventHandlerMethodInfo = currentType.GetMethod("CompleteEventHandler");
                }
                return completeEventHandlerMethodInfo;
            }
        }

        private static Tuple<EventInfo, Delegate> GetEventAndHandler(Delegate function)
        {
            var srcMethod = function.Method;
            var methodName = srcMethod.Name;
            if (!methodName.EndsWith("Async"))
                throw new ArgumentException(methodName + " must be ends with Async");

            var target = function.Target;
            var targetType = target.GetType();

            var completeEventName = methodName.Substring(0, methodName.Length - 5) + "Completed";
            EventInfo eventInfo = targetType.GetEvent(completeEventName);

            var eventArgsName = completeEventName + "EventArgs";
            Type eventArgsType = targetType.Assembly.GetType(targetType.Namespace + "." + eventArgsName);

            var handlerMethod = CompleteEventHandlerMethodInfo.MakeGenericMethod(eventArgsType);

            Type handlerType = Type.GetType(String.Format("System.EventHandler`1[[{0},{1}]]", eventArgsType.FullName, AssemblyName));
            Delegate handler = Delegate.CreateDelegate(handlerType, handlerMethod);

            return new Tuple<EventInfo, Delegate>(eventInfo, handler);
        }

        private static MethodInfo GetMethodWithUserState(Delegate function)
        {
            var paramTypes = function.Method.GetParameters().Select(info => info.ParameterType).Concat(new Type[] { typeof(object) }).ToArray();
            return function.Target.GetType().GetMethod(function.Method.Name, paramTypes);
        }

        public static void CompleteEventHandler<TArgs>(object sender, TArgs eventArgs)
            where TArgs : AsyncCompletedEventArgs
        {
            if (eventArgs.Cancelled) return;
            if (eventArgs.Error != null)
            {
                if (eventArgs.UserState is Action<Exception>)
                    (eventArgs.UserState as Action<Exception>)(eventArgs.Error);
                else if (eventArgs.UserState is Action<object, Exception>)
                    (eventArgs.UserState as Action<object, Exception>)(null, eventArgs.Error);
                else
                    throw eventArgs.Error;
            }
            else
            {
                if (eventArgs.UserState is Action<Exception>)
                    (eventArgs.UserState as Action<Exception>)(null);
                else if (eventArgs.UserState is Action<object, Exception>)
                {
                    object result = eventArgs.GetType().GetProperty("Result").GetValue(eventArgs, null);
                    (eventArgs.UserState as Action<object, Exception>)(result, null);
                }
            }
        }

        private T serviceClient;

        public ServiceProxy(T serviceClient)
        {
            this.serviceClient = serviceClient;
        }

        public T ServiceClient
        {
            get
            {
                return this.serviceClient;
            }
        }

        private Dictionary<Delegate, MethodInfo> functionCache;

        public Dictionary<Delegate, MethodInfo> FunctionCache
        {
            get
            {
                if (functionCache == null)
                    functionCache = new Dictionary<Delegate, MethodInfo>();
                return functionCache;
            }
        }

        protected MethodInfo GetFunc(Delegate function)
        {
            if (!function.Target.Equals(this.ServiceClient))
                throw new Exception("must use ServiceClient object");

            if (!this.FunctionCache.ContainsKey(function))
            {
                var func = GetMethodWithUserState(function);
                var args = GetEventAndHandler(function);
                args.Item1.AddEventHandler(function.Target, args.Item2);
                FunctionCache.Add(function, func);
            }

            return FunctionCache[function];
        }

        public void Reset(T serviceClient)
        {
            this.serviceClient = serviceClient;
            if (this.functionCache != null)
                this.functionCache.Clear();
        }

        public void PerformFunction(Action function, Action<Exception> callback)
        {
            var func = GetFunc(function);
            var userState = callback;
            func.Invoke(function.Target, new object[] { userState });
        }

        public void PerformFunction<TArg>(Action<TArg> function, TArg arg, Action<Exception> callback)
        {
            var func = GetFunc(function);
            var userState = callback;
            func.Invoke(function.Target, new object[] { arg, userState });
        }

        public void PerformFunction<TArg1, TArg2>(Action<TArg1, TArg2> function, TArg1 arg1, TArg2 arg2, Action<Exception> callback)
        {
            var func = GetFunc(function);
            var userState = callback;
            func.Invoke(function.Target, new object[] { arg1, arg2, userState });
        }

        public void PerformFunction<TArg1, TArg2, TArg3>(Action<TArg1, TArg2, TArg3> function, TArg1 arg1, TArg2 arg2, TArg3 arg3, Action<Exception> callback)
        {
            var func = GetFunc(function);
            var userState = callback;
            func.Invoke(function.Target, new object[] { arg1, arg2, arg3, userState });
        }

        public void PerformFunction<TArg1, TArg2, TArg3, TArg4>(Action<TArg1, TArg2, TArg3, TArg4> function, TArg1 arg1, TArg2 arg2, TArg3 arg3, TArg4 arg4, Action<Exception> callback)
        {
            var func = GetFunc(function);
            var userState = callback;
            func.Invoke(function.Target, new object[] { arg1, arg2, arg3, arg4, userState });
        }

        public void PerformFunction<TResult>(Action function, Action<TResult, Exception> callback)
        {
            var func = GetFunc(function);
            Action<object, Exception> userState = (obj, error) =>
            {
                if (error == null)
                    callback((TResult)obj, null);
                else
                    callback(default(TResult), error);
            };
            func.Invoke(function.Target, new object[] { userState });
        }

        public void PerformFunction<TArg, TResult>(Action<TArg> function, TArg arg, Action<TResult, Exception> callback)
        {
            var func = GetFunc(function);
            Action<object, Exception> userState = (obj, error) =>
            {
                if (error == null)
                    callback((TResult)obj, null);
                else
                    callback(default(TResult), error);
            };
            func.Invoke(function.Target, new object[] { arg, userState });
        }

        public void PerformFunction<TArg1, TArg2, TResult>(Action<TArg1, TArg2> function, TArg1 arg1, TArg2 arg2, Action<TResult, Exception> callback)
        {
            var func = GetFunc(function);
            Action<object, Exception> userState = (obj, error) =>
            {
                if (error == null)
                    callback((TResult)obj, null);
                else
                    callback(default(TResult), error);
            };
            func.Invoke(function.Target, new object[] { arg1, arg2, userState });
        }

        public void PerformFunction<TArg1, TArg2, TArg3, TResult>(Action<TArg1, TArg2, TArg3> function, TArg1 arg1, TArg2 arg2, TArg3 arg3, Action<TResult, Exception> callback)
        {
            var func = GetFunc(function);
            Action<object, Exception> userState = (obj, error) =>
            {
                if (error == null)
                    callback((TResult)obj, null);
                else
                    callback(default(TResult), error);
            };
            func.Invoke(function.Target, new object[] { arg1, arg2, arg3, userState });
        }

        public void PerformFunction<TArg1, TArg2, TArg3, TArg4, TResult>(Action<TArg1, TArg2, TArg3, TArg4> function, TArg1 arg1, TArg2 arg2, TArg3 arg3, TArg4 arg4, Action<TResult, Exception> callback)
        {
            var func = GetFunc(function);
            Action<object, Exception> userState = (obj, error) =>
            {
                if (error == null)
                    callback((TResult)obj, null);
                else
                    callback(default(TResult), error);
            };
            func.Invoke(function.Target, new object[] { arg1, arg2, arg3, arg4, userState });
        }
    }
}
